from csc148_queue import Queue


class Tree:
    """
    A bare-bones Tree ADT that identifies the root with the entire tree.
    """

    def __init__(self, value=None, children=None):
        """
        Create Tree self with content value and 0 or more children

        :param value: value contained in this tree
        :type value: object
        :param children: possibly-empty list of children
        :type children: list[Tree]
        """
        self.value = value
        # copy children if not None
        self.children = children.copy() if children else []

    def __repr__(self):
        """
        Return representation of Tree (self) as string that
        can be evaluated into an equivalent Tree.

        :rtype: str

        >>> t1 = Tree(5)
        >>> t1
        Tree(5)
        >>> t2 = Tree(7, [t1])
        >>> t2
        Tree(7, [Tree(5)])
        """
        # Our __repr__ is recursive, because it can also be called
        # via repr...!
        return ('Tree({}, {})'.format(repr(self.value), repr(self.children))
                if self.children
                else 'Tree({})'.format(repr(self.value)))

    def __eq__(self, other):
        """
        Return whether this Tree is equivalent to other.

        :param other: object to compare to self
        :type other: object}Tree
        :rtype: bool

        >>> t1 = Tree(5)
        >>> t2 = Tree(5, [])
        >>> t1 == t2
        True
        >>> t3 = Tree(5, [t1])
        >>> t2 == t3
        False
        """
        return (type(self) is type(other) and
                self.value == other.value and
                self.children == other.children)

    def __str__(self, indent=0):
        """
        Produce a user-friendly string representation of Tree self,
        indenting each level as a visual clue.

        :param indent: amount to indent each level of tree
        :type indent: int
        :rtype: str

        >>> t = Tree(17)
        >>> print(t)
        17
        >>> t1 = Tree(19, [t, Tree(23)])
        >>> print(t1)
        19
           17
           23
        >>> t3 = Tree(29, [Tree(31), t1])
        >>> print(t3)
        29
           31
           19
              17
              23
        """
        root_str = indent * " " + str(self.value)
        return '\n'.join([root_str] +
                         [c.__str__(indent + 3) for c in self.children])

    def __contains__(self, v):
        """
        Return whether Tree self contains v.

        :param v: value to search this tree for
        :type v: object

        >>> t = Tree(17)
        >>> t.__contains__(17)
        True
        >>> t = descendants_from_list(Tree(19), [1, 2, 3, 4, 5, 6, 7], 3)
        >>> t.__contains__(5)
        True
        >>> t.__contains__(18)
        False
        """
        if len(self.children) == 0:
            # self is a leaf
            return self.value == v
        else:
            # self is not a leaf
            return self.value == v or any([v in x for x in self.children])


def leaf_count(t):
    """
    Return the number of leaves in Tree t.

    :param t: tree to count the leaves of
    :type t: Tree
    :rtype: int

    >>> t = Tree(7)
    >>> leaf_count(t)
    1
    >>> t = descendants_from_list(Tree(7), [0, 1, 3, 5, 7, 9, 11, 13], 3)
    >>> leaf_count(t)
    6
    """
    if len(t.children) == 0:
        # t is a leaf
        return 1
    else:
        # t is an internal node
        return sum([leaf_count(c) for c in t.children])


def list_internal(t):
    """
    Return list of values in internal nodes of t.

    :param t: tree to list internal values of
    :type t: Tree
    :rtype: list[object]

    >>> t = Tree(0)
    >>> list_internal(t)
    []
    >>> t = descendants_from_list(Tree(0), [1, 2, 3, 4, 5, 6, 7, 8], 3)
    >>> L = list_internal(t)
    >>> L.sort()
    >>> L
    [0, 1, 2]
    """
    if len(t.children) == 0:
        return []
    else:
        return [t.value] + gather_lists([list_internal(c) for c in t.children])


def arity(t):
    """
    Return the maximum branching factor (arity) of Tree t.

    :param t: tree to find the arity of
    :type t: Tree
    :rtype: int

    >>> t = Tree(23)
    >>> arity(t)
    0
    >>> tn2 = Tree(2, [Tree(4), Tree(4.5), Tree(5), Tree(5.75)])
    >>> tn3 = Tree(3, [Tree(6), Tree(7)])
    >>> tn1 = Tree(1, [tn2, tn3])
    >>> arity(tn1)
    4
    """
    return max([len(t.children)] + [arity(n) for n in t.children])


# helper function that may be useful in the functions
# above
def gather_lists(list_):
    """
    Concatenate all the sublists of L and return the result.

    :param list_: list of lists to concatenate
    :type list_: list[list[object]]
    :rtype: list[object]

    >>> gather_lists([[1, 2], [3, 4, 5]])
    [1, 2, 3, 4, 5]
    >>> gather_lists([[6, 7], [8], [9, 10, 11]])
    [6, 7, 8, 9, 10, 11]
    """
    new_list = []
    for l in list_:
        new_list += l
    return new_list


# helpful helper function
def descendants_from_list(t, list_, arity):
    """
    Populate Tree t's descendants from list_, filling them
    in in level order, with up to arity children per node.
    Then return t.

    :param t: tree to populate from list_
    :type t: Tree
    :param list_: list of values to populate from
    :type list_: list
    :param arity: maximum branching factor
    :type arity: int
    :rtype: Tree

    >>> descendants_from_list(Tree(0), [1, 2, 3, 4], 2)
    Tree(0, [Tree(1, [Tree(3), Tree(4)]), Tree(2)])
    """
    q = Queue()
    q.add(t)
    list_ = list_.copy()
    while not q.is_empty():  # unlikely to happen
        new_t = q.remove()
        for i in range(0, arity):
            if len(list_) == 0:
                return t  # our work here is done
            else:
                new_t_child = Tree(list_.pop(0))
                new_t.children.append(new_t_child)
                q.add(new_t_child)
    return t


def is_full(t, n):
    """
     Return whether tree t is full with respect to arity n

    :param t: a general tree
    :type t: Tree
    :param n: a given branching factor for the tree
    :type n: int
    :rtype: bool

    >>> t1 = descendants_from_list(Tree(0), [], 3)
    >>> is_full(t1, 3)
    True
    >>> t2 = descendants_from_list(Tree(0), [1, 2, 3], 3)
    >>> is_full(t2, 3)
    True
    >>> t3 = descendants_from_list(Tree(0), [1, 2, 3, 4], 3)
    >>> is_full(t3, 3)
    False
    >>> t4 = descendants_from_list(Tree(0), [1, 2, 3, 4, 5, 6], 3)
    >>> is_full(t4, 3)
    True
    """
    # to be developed by you

    if len(t.children) == 0:
        return True
    elif len(t.children) != n:
        return False
    else:
        return all([is_full(c,n) for c in t.children])


def is_full1(t, n):
    """
     Return whether tree t is full with respect to arity n

    :param t: a general tree
    :type t: Tree
    :param n: a given branching factor for the tree
    :type n: int
    :rtype: bool

    >>> t1 = descendants_from_list(Tree(0), [], 3)
    >>> is_full1(t1, 3)
    True
    >>> t2 = descendants_from_list(Tree(0), [1, 2, 3], 3)
    >>> is_full1(t2, 3)
    True
    >>> t3 = descendants_from_list(Tree(0), [1, 2, 3, 4], 3)
    >>> is_full1(t3, 3)
    False
    >>> t4 = descendants_from_list(Tree(0), [1, 2, 3, 4, 5, 6], 3)
    >>> is_full1(t4, 3)
    True
    """
    if len(t.children) == 0:
        return True
    elif len(t.children) != n:
        return False
    else:
        return all([is_full1(c,n) for c in t.children])


def is_full2(t, n):
    """
     Return whether tree t is full with respect to arity n

    :param t: a general tree
    :type t: Tree
    :param n: a given branching factor for the tree
    :type n: int
    :rtype: bool

    >>> t1 = descendants_from_list(Tree(0), [], 3)
    >>> is_full2(t1, 3)
    True
    >>> t2 = descendants_from_list(Tree(0), [1, 2, 3], 3)
    >>> is_full2(t2, 3)
    True
    >>> t3 = descendants_from_list(Tree(0), [1, 2, 3, 4], 3)
    >>> is_full2(t3, 3)
    False
    >>> t4 = descendants_from_list(Tree(0), [1, 2, 3, 4, 5, 6], 3)
    >>> is_full2(t4, 3)
    True
    """
    if len(t.children) == 0:
        return True
    elif len(t.children) != n:
        return False
    else:
        i = 0
        while i < len(t.children) and is_full2(t.children[i], n):
            i +=1
        if i == len(t.children):
            return True
        else:
            return False


def is_full3(t, n):
    """
     Return whether tree t is full with respect to arity n

    :param t: a general tree
    :type t: Tree
    :param n: a given branching factor for the tree
    :type n: int
    :rtype: bool

    >>> t1 = descendants_from_list(Tree(0), [], 3)
    >>> is_full3(t1, 3)
    True
    >>> t2 = descendants_from_list(Tree(0), [1, 2, 3], 3)
    >>> is_full3(t2, 3)
    True
    >>> t3 = descendants_from_list(Tree(0), [1, 2, 3, 4], 3)
    >>> is_full3(t3, 3)
    False
    >>> t4 = descendants_from_list(Tree(0), [1, 2, 3, 4, 5, 6], 3)
    >>> is_full3(t4, 3)
    True
    """
    q = Queue()
    q.add(t)

    while not q.is_empty():
        t1 = q.remove()
        if len(t1.children) == n:
            for i in range(len(t1.children)):
                q.add(t1.children[i])
        elif len(t1.children) == 0:
            pass
        else: return False
    return True


if __name__ == '__main__':
    import doctest
    doctest.testmod()
